﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace WindowsFormsApplication1
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            Icon = SystemIcons.Application;
        }

        private void InitTimer_Tick(object sender, EventArgs e)
        {
            InitTimer.Enabled = false;
            init();
        }

        private uint num_total_dirs_ = 0, num_total_files_ = 0;
        private long total_bytes_ = 0;
        private uint num_found_dirs_ = 0, num_found_files_ = 0;
        private bool params_characters_enabled_;
        private uint params_max_length_;

        private void init()
        {
            try
            {
                FolderBrowserDialog dialog = new FolderBrowserDialog();
                dialog.RootFolder = Environment.SpecialFolder.MyComputer;
                dialog.Description = "Wybierz katalog";
                dialog.ShowNewFolderButton = false;
                if (dialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
                {
                    ParamsForm params_form = new ParamsForm();
                    if (params_form.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
                    {
                        params_characters_enabled_ = params_form.characters_enabled_;
                        params_max_length_ = params_form.max_length_;

                        Cursor = Cursors.WaitCursor;
                        main_list_view_.BeginUpdate();
                        process_directory(dialog.SelectedPath);

                        sort_list();
                        main_list_view_.EndUpdate();
                        Cursor = Cursors.Default;

                        update_status_bar();
                    }
                    else
                        Close();
                }
                else
                    Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Błąd", MessageBoxButtons.OK, MessageBoxIcon.Error);
                Close();
            }
        }

        private void sort_list()
        {
            main_list_view_.ListViewItemSorter = new ListItemSorter();
            main_list_view_.ListViewItemSorter = null;
            //main_list_view_.Sort(); // Not necessary.
        }

        private void update_status_bar()
        {
            status_label_.Text = String.Format("Wszystkich katalogów {0} plików {1}, razem {2}, znalezionych katalogów {3} plików {4}",
                num_total_dirs_,
                num_total_files_,
                size_to_string(total_bytes_),
                num_found_dirs_,
                num_found_files_);
        }

        private void process_directory(string path)
        {
            DirectoryInfo dir_info = new DirectoryInfo(path);
            process_directory(dir_info);
        }

        private long process_directory(DirectoryInfo directory_info)
        {
            FileSystemInfo[] sub_infos = directory_info.GetFileSystemInfos();
            long sum_size = 0;
            foreach (FileSystemInfo sub_info in sub_infos)
            {
                bool is_dir = (sub_info.Attributes & FileAttributes.Directory) != 0;
                long sub_size;
                if (is_dir)
                {
                    ++num_total_dirs_;
                    sub_size = process_directory((DirectoryInfo)sub_info);
                }
                else
                {
                    ++num_total_files_;
                    sub_size = ((FileInfo)sub_info).Length;
                    total_bytes_ += sub_size;
                }

                sum_size += sub_size;

                if (!name_is_valid(sub_info.Name))
                {
                    if (is_dir)
                        ++num_found_dirs_;
                    else
                        ++num_found_files_;

                    Tag tag = new Tag();
                    tag.file_system_info_ = sub_info;
                    tag.size_ = size_to_string(sub_size);

                    ListViewItem item = new ListViewItem();
                    main_list_view_.Items.Add(item);
                    item.Tag = tag;
                    tag_to_item_parameters(tag, item);
                }
            }
            return sum_size;
        }

        private bool name_is_valid(string name)
        {
            if (params_characters_enabled_)
            {
                foreach (char ch in name)
                {
                    ushort code = (ushort)ch;
                    if (code < 32 || code >= 128)
                        return false;
                }
            }
            
            if (params_max_length_ > 0)
            {
                if (Path.GetFileNameWithoutExtension(name).Length > params_max_length_)
                    return false;
            }
            
            return true;
        }

        private string size_to_string(long bytes)
        {
            if (bytes >= 1073741824)
                return ((double)bytes / 1073741824.0).ToString("F2") + " GB ";
            if (bytes >= 1048576)
                return ((double)bytes / 1048576.0).ToString("F2") + " MB";
            if (bytes >= 1024)
                return ((double)bytes / 1024.0).ToString("F2") + " KB";
            return bytes.ToString() + " B";
        }

        private void rename_button__Click(object sender, EventArgs e)
        {
            if (main_list_view_.SelectedItems.Count == 1)
                main_list_view_.SelectedItems[0].BeginEdit();
        }

        private void main_list_view__AfterLabelEdit(object sender, LabelEditEventArgs e)
        {
            if (e.CancelEdit)
                return;
            if (string.IsNullOrEmpty(e.Label))
            {
                e.CancelEdit = true;
                return;
            }
            ListViewItem item = main_list_view_.Items[e.Item];
            Tag tag = (Tag)item.Tag;
            if (e.Label == tag.file_system_info_.Name)
            {
                e.CancelEdit = true;
                return;
            }

            try
            {
                string new_path = Path.Combine(Path.GetDirectoryName(tag.file_system_info_.FullName), e.Label);
                if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                    Directory.Move(tag.file_system_info_.FullName, new_path);
                else
                    File.Move(tag.file_system_info_.FullName, new_path);

                apply_rename(new Tag[] { tag }, new string[] { new_path });
            }
            catch (Exception ex)
            {
                e.CancelEdit = true;
                MessageBox.Show(this, ex.Message, "Zmień nazwę", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void apply_rename(Tag[] tags, string[] new_paths)
        {
            if (tags.Length == 0) return;

            // Super sophostocated algorithm :)

            main_list_view_.BeginUpdate();
            int tag_index = 0, tag_count = tags.Length;
            for (int item_index = 0, item_count = main_list_view_.Items.Count; item_index < item_count; ++item_index)
            {
                ListViewItem item = main_list_view_.Items[item_index];
                Tag tag = (Tag)item.Tag;
                if (tag == tags[tag_index])
                {
                    string old_path = tag.file_system_info_.FullName;
                    string old_path_backslash = old_path + @"\";
                    int old_path_len = old_path.Length;
                    
                    apply_rename_to_item(item, new_paths[tag_index]);
                    
                    for (int j = item_index + 1; j < item_count; ++j)
                    {
                        ListViewItem item_j = main_list_view_.Items[j];
                        Tag tag_j = (Tag)item_j.Tag;
                        if (tag_j.file_system_info_.FullName.Length > old_path_len
                            && tag_j.file_system_info_.FullName.Substring(0, old_path_len + 1) == old_path_backslash)
                        {
                            apply_rename_to_item(item_j, new_paths[tag_index] + tag_j.file_system_info_.FullName.Substring(old_path_len));
                        }
                        //else
                        //    break;
                    }

                    for (int tag_j = tag_index + 1; tag_j < tag_count; ++tag_j)
                    {
                        if (new_paths[tag_j].Length > old_path_len
                            && new_paths[tag_j].Substring(0, old_path_len + 1) == old_path_backslash)
                        {
                            new_paths[tag_j] = new_paths[tag_index] + new_paths[tag_j].Substring(old_path_len);
                        }
                    }

                    if (++tag_index == tag_count)
                        break;
                }
            }
            main_list_view_.EndUpdate();
        }

        private void apply_delete(Tag[] tags)
        {
            if (tags.Length == 0) return;

            // Super sophostocated algorithm :)

            main_list_view_.BeginUpdate();
            int tag_index = 0, tag_count = tags.Length;
            for (int item_index = 0, item_count = main_list_view_.Items.Count; item_index < item_count; )
            {
                ListViewItem item = main_list_view_.Items[item_index];
                Tag tag = (Tag)item.Tag;
                if (tag == tags[tag_index])
                {
                    string old_path = tag.file_system_info_.FullName;
                    string old_path_backslash = old_path + @"\";
                    int old_path_len = old_path.Length;

                    main_list_view_.Items.RemoveAt(item_index);
                    --item_count;

                    for (int j = item_index; j < item_count; )
                    {
                        ListViewItem item_j = main_list_view_.Items[j];
                        Tag tag_j = (Tag)item_j.Tag;
                        if (tag_j.file_system_info_.FullName.Length > old_path_len
                            && tag_j.file_system_info_.FullName.Substring(0, old_path_len + 1) == old_path_backslash)
                        {
                            main_list_view_.Items.RemoveAt(j);
                            --item_count;
                        }
                        else
                            ++j;
                    }

                    for (int tag_j = tag_index + 1; tag_j < tag_count; )
                    {
                        if (tags[tag_j].file_system_info_.FullName.Length > old_path_len
                            && tags[tag_j].file_system_info_.FullName.Substring(0, old_path_len + 1) == old_path_backslash)
                        {
                            --tag_count;
                            for (int k = tag_j; k < tag_count; ++k)
                                tags[k] = tags[k + 1];
                        }
                        else
                            ++tag_j;
                    }

                    if (++tag_index == tag_count)
                        break;
                }
                else
                    ++item_index;
            }
            main_list_view_.EndUpdate();
        }

        private void apply_rename_to_item(ListViewItem item, string new_path)
        {
            Tag tag = (Tag)item.Tag;

            if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                tag.file_system_info_ = new DirectoryInfo(new_path);
            else
                tag.file_system_info_ = new FileInfo(new_path);

            tag_to_item_parameters(tag, item);
        }

        private void tag_to_item_parameters(Tag tag, ListViewItem item)
        {
            item.SubItems.Clear();
            item.Text = tag.file_system_info_.Name;
            item.SubItems.Add(Path.GetDirectoryName(tag.file_system_info_.FullName));
            item.SubItems.Add(tag.size_);
            item.SubItems.Add(tag.file_system_info_.LastWriteTime.ToString());

            bool is_dir = (tag.file_system_info_.Attributes & FileAttributes.Directory) != 0;
            item.ImageIndex = is_dir ? 0 : 1;
        }

        private void Form1_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.F2:
                    rename_button__Click(null, null);
                    break;

                case Keys.F3:
                    auto_fix_button__Click(null, null);
                    break;

                case Keys.Delete:
                    if (e.Shift)
                        delete_permanently_button__Click(null, null);
                    else
                        delete_button__Click(null, null);
                    break;

                case Keys.A:
                    if (e.Control)
                        select_all_button__Click(null, null);
                    break;
            }
        }

        private void auto_fix_button__Click(object sender, EventArgs e)
        {
            int selected_count = main_list_view_.SelectedItems.Count;
            if (selected_count == 0)
                return;

            AutoFixForm form = new AutoFixForm();

            if (params_characters_enabled_)
            {
                form.russian_checkbox_.Checked = true;
                form.polish_checkbox_.Checked = true;
            }

            if (params_max_length_ > 0)
            {
                form.length_checkbox_.Checked = true;
                form.length_textbox_.Text = params_max_length_.ToString();
            }

            if (form.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                bool polish = form.polish_checkbox_.Checked;
                bool russian = form.russian_checkbox_.Checked;
                bool length = form.length_checkbox_.Checked;

                if (polish || russian || length)
                {
                    Tag[] tags = new Tag[selected_count];
                    for (int i = 0; i < selected_count; ++i)
                        tags[i] = (Tag)main_list_view_.SelectedItems[i].Tag;

                    Transcription trans = null;
                    if (polish || russian)
                    {
                        trans = new Transcription();
                        if (russian)
                            trans.add_transcription(Transcription.russian_transcriptions_[form.russian_combobox_.SelectedIndex]);
                        if (polish)
                            trans.add_transcription(Transcription.polish_transcription_);
                    }

                    Cursor = Cursors.WaitCursor;
                    auto_fix(tags, trans, form.max_length_);
                    Cursor = Cursors.Default;
                }
            }
        }

        // transcription == null - not use transcription
        // max_length == 0 - not use max length
        private void auto_fix(Tag[] tags, Transcription transcription, uint max_length)
        {
            List<Tag> applied_tags = new List<Tag>();
            List<string> applied_new_paths = new List<string>();

            for (int i = tags.Length; i-- != 0; )
            {
                Tag tag = tags[i];
                bool retry = false;
                do
                {
                    string old_path = tag.file_system_info_.FullName;
                    string dir = Path.GetDirectoryName(old_path);
                    string old_name = tag.file_system_info_.Name;

                    string new_name;
                    if (transcription != null)
                        new_name = transcription.transcribe(old_name);
                    else
                        new_name = old_name;
                    if (max_length > 0)
                    {
                        string name_no_ext = Path.GetFileNameWithoutExtension(new_name);
                        if (name_no_ext.Length > max_length)
                            new_name = new_name.Substring(0, (int)max_length) + new_name.Substring(name_no_ext.Length);
                    }

                    if (new_name != old_name)
                    {
                        string new_path = Path.Combine(dir, new_name);

                        try
                        {
                            if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                                Directory.Move(old_path, new_path);
                            else
                                File.Move(old_path, new_path);

                            applied_tags.Add(tag);
                            applied_new_paths.Add(new_path);

                            break;
                        }
                        catch (Exception ex)
                        {
                            DialogResult result = MessageBox.Show(this, ex.Message, "Napraw automatycznie", MessageBoxButtons.AbortRetryIgnore, MessageBoxIcon.Error);

                            if (result == System.Windows.Forms.DialogResult.Retry)
                                retry = true;
                            else if (result == System.Windows.Forms.DialogResult.Abort)
                                return;
                            else // Ignore
                                retry = false;
                        }
                    }
                }
                while (retry);
            }

            applied_tags.Reverse();
            applied_new_paths.Reverse();

            apply_rename(applied_tags.ToArray(), applied_new_paths.ToArray());
        }

        private void delete_button__Click(object sender, EventArgs e)
        {
            int selected_count = main_list_view_.SelectedItems.Count;
            if (selected_count == 0)
                return;

            string question;
            if (selected_count == 1)
            {
                Tag tag = (Tag)main_list_view_.SelectedItems[0].Tag;
                question = "Czy na pewno chcesz usunąć zaznaczony ";
                if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                    question += "katalog:\n";
                else
                    question += "plik:\n";
                question += tag.file_system_info_.FullName;
            }
            else
                question = string.Format("Czy na pewno chcesz usunąć {0} zaznaczonych obiektów ?", selected_count);

            if (MessageBox.Show(this, question, "Usuń do kosza", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == System.Windows.Forms.DialogResult.Yes)
            {
                Tag[] tags_to_delete = new Tag[selected_count];
                for (int i = 0; i < selected_count; ++i)
                    tags_to_delete[i] = (Tag)main_list_view_.SelectedItems[i].Tag;

                Cursor = Cursors.WaitCursor;
                delete_to_recycle_bin(tags_to_delete);
                Cursor = Cursors.Default;
            }
        }

        private void delete_permanently_button__Click(object sender, EventArgs e)
        {
            int selected_count = main_list_view_.SelectedItems.Count;
            if (selected_count == 0)
                return;

            string question;
            if (selected_count == 1)
            {
                Tag tag = (Tag)main_list_view_.SelectedItems[0].Tag;
                question = "Czy na pewno chcesz bezpowrotnie usunąć zaznaczony ";
                if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                    question += "katalog:\n";
                else
                    question += "plik:\n";
                question += tag.file_system_info_.FullName;
            }
            else
                question = string.Format("Czy na pewno chcesz bezpowrotnie usunąć {0} zaznaczonych obiektów ?", selected_count);

            if (MessageBox.Show(this, question, "Usuń bezpowrotnie", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == System.Windows.Forms.DialogResult.Yes)
            {
                Tag[] tags_to_delete = new Tag[selected_count];
                for (int i = 0; i < selected_count; ++i)
                    tags_to_delete[i] = (Tag)main_list_view_.SelectedItems[i].Tag;

                Cursor = Cursors.WaitCursor;
                delete_permanently(tags_to_delete);
                Cursor = Cursors.Default;
            }
        }

        private void delete_permanently(Tag[] tags)
        {
            List<Tag> applied_tags = new List<Tag>();

            for (int i = tags.Length; i-- != 0; )
            {
                Tag tag = tags[i];
                bool retry = false;
                do
                {
                    try
                    {
                        if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                            Directory.Delete(tag.file_system_info_.FullName, true);
                        else
                            File.Delete(tag.file_system_info_.FullName);
                        applied_tags.Add(tag);
                        break;
                    }
                    catch (Exception ex)
                    {
                        DialogResult result = MessageBox.Show(this, ex.Message, "Usuń bezpowrotnie", MessageBoxButtons.AbortRetryIgnore, MessageBoxIcon.Error);
                        
                        if (result == System.Windows.Forms.DialogResult.Retry)
                            retry = true;
                        else if (result == System.Windows.Forms.DialogResult.Abort)
                            return;
                        else // Ignore
                            retry = false;
                    }
                }
                while (retry);
            }

            applied_tags.Reverse();
            apply_delete(applied_tags.ToArray());
        }

        private void delete_to_recycle_bin(Tag[] tags)
        {
            List<Tag> applied_tags = new List<Tag>();

            for (int i = tags.Length; i-- != 0; )
            {
                Tag tag = tags[i];
                bool retry = false;
                do
                {
                    try
                    {
                        if ((tag.file_system_info_.Attributes & FileAttributes.Directory) != 0)
                        {
                            Microsoft.VisualBasic.FileIO.FileSystem.DeleteDirectory(
                                tag.file_system_info_.FullName,
                                Microsoft.VisualBasic.FileIO.UIOption.OnlyErrorDialogs,
                                Microsoft.VisualBasic.FileIO.RecycleOption.SendToRecycleBin,
                                Microsoft.VisualBasic.FileIO.UICancelOption.DoNothing);
                        }
                        else
                        {
                            Microsoft.VisualBasic.FileIO.FileSystem.DeleteFile(
                                tag.file_system_info_.FullName,
                                Microsoft.VisualBasic.FileIO.UIOption.OnlyErrorDialogs,
                                Microsoft.VisualBasic.FileIO.RecycleOption.SendToRecycleBin,
                                Microsoft.VisualBasic.FileIO.UICancelOption.DoNothing);
                        }

                        applied_tags.Add(tag);
                        break;
                    }
                    catch (Exception ex)
                    {
                        DialogResult result = MessageBox.Show(this, ex.Message, "Usuń do kosza", MessageBoxButtons.AbortRetryIgnore, MessageBoxIcon.Error);

                        if (result == System.Windows.Forms.DialogResult.Retry)
                            retry = true;
                        else if (result == System.Windows.Forms.DialogResult.Abort)
                            return;
                        else // Ignore
                            retry = false;
                    }
                }
                while (retry);
            }

            applied_tags.Reverse();
            apply_delete(applied_tags.ToArray());
        }

        private void main_list_view__ItemActivate(object sender, EventArgs e)
        {
            ListViewItem item = main_list_view_.FocusedItem;
            if (item == null) return;
            Tag tag = (Tag)item.Tag;
            try
            {
                System.Diagnostics.Process.Start(tag.file_system_info_.FullName);
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Otwórz", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void open_button__Click(object sender, EventArgs e)
        {
            main_list_view__ItemActivate(null, null);
        }

        private void select_all_button__Click(object sender, EventArgs e)
        {
            main_list_view_.BeginUpdate();
            foreach (ListViewItem item in main_list_view_.Items)
                item.Selected = true;
            main_list_view_.EndUpdate();
        }

        private void remove_from_list_button__Click(object sender, EventArgs e)
        {
            main_list_view_.BeginUpdate();
            while (main_list_view_.SelectedItems.Count > 0)
                main_list_view_.SelectedItems[0].Remove();
            main_list_view_.EndUpdate();
        }
    }

    class ListItemSorter : System.Collections.IComparer
    {
        public int Compare(object lhs, object rhs)
        {
            ListViewItem lhs_item = (ListViewItem)lhs;
            ListViewItem rhs_item = (ListViewItem)rhs;
            Tag lhs_tag = (Tag)lhs_item.Tag;
            Tag rhs_tag = (Tag)rhs_item.Tag;
            return string.Compare(lhs_tag.file_system_info_.FullName, rhs_tag.file_system_info_.FullName, true);
        }
    };
}
